Program EMSTest;

{A simple program to test the functionality of the lo-tech 2MB EMS
 board. This program accessess the hardware directly; there is no
 dependence on EMS services (int 67h) being available.
 
 Revision r002.
 Written by James Pearce, February 2014.
 Updated May-14.}

uses dos, crt;

CONST
  PAGESIZE  : WORD = 16384; {16KB pages}
  MINPAGES  : WORD = 32;    {min config is 512KB, = 32 x 16KB pages}
  MAXPAGES  : WORD = 128;   {max config is 2MB, = 128 x 16KB pages}

  FUNC_OK   : INTEGER = 1;
  FUNC_ERR  : INTEGER = -1;

  EMSPage   : BYTE = 1;

VAR
  TOTALPAGES : WORD;
  PageMap : Array[0..255] of BYTE;

  REG0, REG1, REG2, REG3     : Word;
  PAGE0, PAGE1, PAGE2, PAGE3 : Word;


{----------------------------------------------------------------------------}
{ Helper functions                                                           }  
{----------------------------------------------------------------------------}


Function COMPSW( source, destination : pointer; words : word) : Word;
{implements COMPSW function, comparing source with destination on a
 word-by-word basis, returning zero if they were OK}
var
  rDS, rSI,
  rES, rDI	  :  Word;
  BlockBad	  :  Word;
begin
  rDS := Seg(source^);
  rSI := Ofs(source^);
  rES := Seg(destination^);
  rDI := Ofs(destination^);
  
  asm
    push	ds
    push	si
    push	es
    push	di
       
    mov		ax, rDS
    mov		ds, ax
    mov		ax, rSI
    mov		si, ax
    
    mov		ax, rES
    mov		es, ax
    mov		ax, rDI
    mov		di, rDI
    
    mov		cx, words
    cld
    
    rep		cmpsw
    jz		@NoDiffs
    
    inc		cx
    
  @NoDiffs:
    mov		BlockBad, cx
    
    pop	di
    pop	es
    pop	si
    pop	ds
  end;{asm}

  COMPSW := BlockBad;
end;{function COMPSW}



function Hex2Word(s : string) : word;
{converts i.e. D000 in ASCII to the corresponding word value}
var
  i             : byte;
  len           : word;
  result        : word;
  digit         : byte;

begin
  len := length(s);
  if len > 4 then len := 4;
  result := 0;
  if len > 0 then begin
    for i := 1 to len do
    begin
      if UpCase(s[i]) <> 'H' then
      begin
        digit := BYTE (s[i]) - 48;
        if digit > 9 then digit := digit - 7;
        result := digit + (result SHL 4);
      end;{if}
    end;{for}
  end;{if}
  Hex2Word := result;
end;{function}


function InHex(value : word) : string;
var
  i, num      :   byte;
  Ch          :   char;

begin
  InHex := '0x';
  InHex[0] := char(6); {set length to 6 chars}
  for i := 1 to 4 do begin
    num := value AND 15;   {lower 4 bits}
    value := value SHR 4; {strip off lower for, for the next iteration}
    if num < 10 then ch := char(num+48) {'0' is ASCII 48}
    else ch := char(num+55); {'A' is ASCII 65}
    InHex[(7-i)] := ch;
  end;{for}
end;{fucntion}

{----------------------------------------------------------------------------}
{ Program specific code                                                      }  
{----------------------------------------------------------------------------}


procedure InitPageMap;
var i : integer;
begin
  for i := 0 to (MAXPAGES-1) do
    PageMap[i] := NOT EMSPage;
end;{procedure}


procedure DisplayPageMap;
{writes the page map in a vaguely usable way, on the screen}
var
  CurrentPage, PagesThisChip, c, i : integer;

begin
  CurrentPage := 0;
  for c := 0 to 3 do
  begin
    PagesThisChip := 0;
    Write('SRAM',c,' (IC',(c+12),'): ');
    for i := 0 to 31 do
    begin
      if (PageMap[CurrentPage] = EMSPage) then
      begin
        Write( '1 ');
        inc(PagesThisChip);
      end else Write( '0 ' );
      inc(CurrentPage);
    end;{for i}
    WriteLn; WriteLn( '              ', PagesThisChip, ' accessible EMS pages detected.');
    WriteLn;
  end;{for c}
end;{procedure DisplayPageMap}


function detect_capacity( pageregister,pageaddr : word ) : integer;
{attempts to detect the available RAM capacity of the board
 by writing to each page and seeing whats there.
 The confirmed number of pages are set in global variable
 TOTALPAGES.
 Note: destructive to current contents.}
var
  currentpage,
  foundpages  : word;
  dataval     : ^word;
  done        : Boolean;

begin
  dataval := ptr(pageaddr,0); {look at first word of the page}

  {first, fill all potential pages with the page number}
  currentpage := 0;
  while (currentpage < MAXPAGES) do
  begin
    Port[pageregister] := currentpage; {set page register}
    dataval^ := currentpage;
    inc(currentpage);
  end;{while}

  {next, check what's there and confirm each detected page
   with a further write}
  currentpage := 0; foundpages := 0;
  done := false;
  while (currentpage < MAXPAGES) do
  begin
    Port[pageregister] := currentpage; {set page register}
    if (dataval^ = currentpage) then
    begin
      {value was correct, now read & write another value}
      dataval^ := $5C;
      if (dataval^ = $5C) then PageMap[currentpage] := EMSPage;
      inc(foundpages);
    end;
    inc(currentpage);
  end;{while}

  if (foundpages = 0) then
  begin
    {no valid results were found; no EMS board detected
     at the specified address}
    TOTALPAGES := 0;
    detect_capacity := FUNC_ERR;
  end else begin
    {some pages were detected; update values and return}
    TOTALPAGES := currentpage;
    detect_capacity := FUNC_OK;
  end;
end;{function detect_capacity}


procedure fill_page(pageregister,pageaddr : word; pageno,fillval : byte);
{fills the specified page via the specified pageregister and pageaddr
 with fillval}
var
  data :  ^byte;
  i    :  word;
begin
  Port[pageregister] := pageno;
  data := ptr(pageaddr,0);
  for i := 1 to PAGESIZE do begin
    data^ := fillval;
    inc(data);
  end;
end;{procedure fill_page}


function check_page(pageregister,pageaddr : word; pageno,fillval : byte) : integer;
{tests the contents of the specified page against data previously filled with
 fill_page. Returns FUNC_OK if all data matches, or FUNC_ERR if one or more
 differences were found}
var
  data  :  ^byte;
  i     :  word;
  error :  Boolean;
begin
  Port[pageregister] := pageno;
  data := ptr(pageaddr,0);
  error := false;
  for i := 1 to PAGESIZE do begin
    if data^ <> fillval then 
    begin
      i := PAGESIZE; {end loop}
      error := true;
    end else inc(data);
  end;
  if error then check_page := FUNC_ERR
  else check_page := FUNC_OK;
end;{function check_page}


function test_registers : integer;
{performs sequential alignment tests}
var res : integer;
begin
  {fill two 64KB windows then see what's there}
  {first 64KB}
  fill_page(REG0,PAGE0,0,$00);
  fill_page(REG1,PAGE1,1,$01);
  fill_page(REG2,PAGE2,2,$02);
  fill_page(REG3,PAGE3,3,$03);
  {second 64KB}
  fill_page(REG0,PAGE0,4,$04);
  fill_page(REG1,PAGE1,5,$05);
  fill_page(REG2,PAGE2,6,$06);
  fill_page(REG3,PAGE3,7,$07);

  {now check what's there}
  res := 0;
  res := res + check_page(REG0,PAGE0,0,$00);
  res := res + check_page(REG1,PAGE1,1,$01);
  res := res + check_page(REG2,PAGE2,2,$02);
  res := res + check_page(REG3,PAGE3,3,$03);
  res := res + check_page(REG0,PAGE0,4,$04);
  res := res + check_page(REG1,PAGE1,5,$05);
  res := res + check_page(REG2,PAGE2,6,$06);
  res := res + check_page(REG3,PAGE3,7,$07);

  if (res = 8) then test_registers := FUNC_OK
  else test_registers := FUNC_ERR;
end;{function test_registers}


function test_aliasing : integer;
{performs window aliasing tests}
var res : integer;
begin
  {fill some pages, then access them via different
   windows and check what's there}

  fill_page(REG0,PAGE0,0,$00);
  fill_page(REG1,PAGE1,1,$01);
  fill_page(REG2,PAGE2,2,$02);
  fill_page(REG3,PAGE3,3,$03);

  fill_page(REG0,PAGE0,4,$04);
  fill_page(REG1,PAGE1,5,$05);
  fill_page(REG2,PAGE2,6,$06);
  fill_page(REG3,PAGE3,7,$07);

  {now check what's there}
  res := 0;
  res := res + check_page(REG0,PAGE0,7,$07);
  res := res + check_page(REG1,PAGE1,6,$06);
  res := res + check_page(REG2,PAGE2,5,$05);
  res := res + check_page(REG3,PAGE3,4,$04);
  res := res + check_page(REG0,PAGE0,3,$03);
  res := res + check_page(REG1,PAGE1,2,$02);
  res := res + check_page(REG2,PAGE2,1,$01);
  res := res + check_page(REG3,PAGE3,0,$00);

  if (res = 8) then 
  begin
    {looks OK so far, now set all four page registers to
     the same value and test that}
    Port[REG0] := 3; Port[REG1] := 3;
    Port[REG2] := 3; Port[REG3] := 3;
    if (COMPSW( Ptr(PAGE0,0), Ptr(PAGE2,0), 16384) = 0) then
      inc(res); {should be 9 if all OK here}

    Port[REG0] := 3; Port[REG1] := 3;
    Port[REG2] := 4; Port[REG3] := 4;
    if (COMPSW( Ptr(PAGE0,0), Ptr(PAGE1,0), 8192) = 0) then
      inc(res); {should be 10 if all OK here}
    if (COMPSW( Ptr(PAGE2,0), Ptr(PAGE3,0), 8192) = 0) then
      inc(res); {should be 11 if all OK here}
    if (COMPSW( Ptr(PAGE0,0), Ptr(PAGE2,0), 8192) <> 0) then
      inc(res); {should be 12 if all OK here}
    if (COMPSW( Ptr(PAGE1,0), Ptr(PAGE3,0), 8192) <> 0) then
      inc(res); {should be 13 if all OK here}
  end;{if}

  if (res = 13) then test_aliasing := FUNC_OK
  else test_aliasing := FUNC_ERR;
end;{function test_aliasing}


procedure run_program;
begin
  InitPagemap;
  Write('Detecting... ');
  if (detect_capacity(REG0,PAGE0) = FUNC_OK) then
  begin
    WriteLn('EMS board detected with ',MAXPAGES,' ',(PAGESIZE/1024) :2:0,
	    'KB pages (',(MAXPAGES*(PAGESIZE/1024)):4:0,'KB)');

    DisplayPageMap;

    if (TOTALPAGES < MINPAGES) then
      WriteLn('Warning: Expected at least ',MINPAGES,' pages.');

    WriteLn; Write('Testing 2x 64KB blocks (8 pages)... ');
    if (test_registers = FUNC_OK) then WriteLn('Passed.')
    else WriteLn('FAILED.');

    Write('Testing window aliasing... ' );
    if (test_aliasing = FUNC_OK) then Write('Passed.')
    else WriteLn('FAILED');

    WriteLn;
  end else WriteLn('EMS board not detected.');
end;{procedure}



var
  BaseAddr, PortBaseAddr : Word;


BEGIN {PROGRAM}
  WriteLn('lo-tech 2MB EMS board hardware test utility. See');
  WriteLn('http://www.lo-tech.co.uk/2MB-EMS-Board');
  WriteLn;
  if ParamCount <> 2 then begin
    WriteLn( 'Usage: EMSTEST [PortBaseAddr] [WindowBaseAddr]');
    WriteLn; WriteLn( ' i.e., EMSTEXT 260 D000');
  end else begin
    PortBaseAddr := Hex2Word(ParamStr(1));
    BaseAddr := Hex2Word(ParamStr(2));
    WriteLn('Using port address ', InHex(PortBaseAddr), ' and memory address ', InHex(BaseAddr));
    REG0 := PortBaseAddr;
    REG1 := REG0 + 1;
    REG2 := REG1 + 1;
    REG3 := REG2 + 1;
    PAGE0 := BaseAddr;
    PAGE1 := PAGE0 + $400;
    PAGE2 := PAGE1 + $400;
    PAGE3 := PAGE2 + $400;
    run_program;
  end; {if/else}
END.{PROGRAM}

{----------------------------------------------------------------------------}
{ End of file.                                                               }  
{----------------------------------------------------------------------------}


